/*  design.cc

    Mark Woolrich, Tim Behrens - FMRIB Image Analysis Group

    Copyright (C) 2002 University of Oxford  */

/*  CCOPYRIGHT  */

#include "design.h"
#include "utils/log.h"
#include "miscmaths/miscmaths.h"
#include "newmat.h"
#include "utils/tracer_plus.h"
#include "gsoptions.h"
#include <string>

using namespace Utilities;
using namespace MISCMATHS;
using namespace NEWMAT;

namespace Gs {

  void Design::setup(bool loadcontrasts)
  {
    Tracer_Plus trace("Design::setup");

    dm = read_vest(GsOptions::getInstance().designfile.value());
    //write_ascii_matrix(dm,"dm"); 
    nevs = dm.Ncols();
    ntpts = dm.Nrows();

    covsplit = read_vest(GsOptions::getInstance().covsplitfile.value());
    //write_ascii_matrix(covsplit,"covsplit"); 
    nevs = dm.Ncols();
    ntpts = dm.Nrows();
    ngs = covsplit.Ncols();

    if(nevs == 0 || ntpts == 0)
	throw Exception(string(GsOptions::getInstance().designfile.value() + " is an invalid design matrix file").data());

    if(ngs == 0 || covsplit.Nrows() == 0)
      throw Exception(string(GsOptions::getInstance().covsplitfile.value() + " is an invalid covariance split file").data());

    if(ntpts != covsplit.Nrows())
      {
	cerr << "design matrix has ntpts=" <<  ntpts << endl;
	cerr << "cov split file has ntpts=" << covsplit.Nrows() << endl;
	throw Exception("The ntpts needs to be the same for both");
      }

    // fill gind:
    gind.ReSize(ntpts);
    gind = 0;
    ntptsing.ReSize(ngs);
    ntptsing = 0;

    for(int t = 1; t <= ntpts; t++)
      {
	for(int g = 1; g <= ngs; g++)
	  {
	    if(covsplit(t,g))
	      {
		if(gind(t)==0)
		  {
		    gind(t) = g;
		    ntptsing(g)++;
		  }
		else
		  {
		    cerr << "t=" << t << endl;
		    throw Exception("t belongs to more than one group");
		  }
	      }
	  }
      }
//     cout << "ntptsing=" << ntptsing << endl;
//     cout << "gind=" << gind << endl;

    // load contrasts:
    if(loadcontrasts)
      {
	try
	  {
	    tcontrasts = read_vest(GsOptions::getInstance().tcontrastsfile.value());
	    numTcontrasts = tcontrasts.Nrows();	    
	  }
	catch(Exception exp)
	  {
	    numTcontrasts = 0;
	  }

	// Check contrast matches design matrix
	if(numTcontrasts > 0 && dm.Ncols() != tcontrasts.Ncols())
	  { 
	    cerr << "Num tcontrast cols  = " << tcontrasts.Ncols() << ", design matrix cols = " << dm.Ncols() << endl;
	    throw Exception("size of design matrix does not match t contrasts\n");
	  }

	try
	  {
	    fcontrasts = read_vest(GsOptions::getInstance().fcontrastsfile.value());
	    numFcontrasts = fcontrasts.Nrows();
	  }
	catch(Exception exp)
	  {
	    numFcontrasts = 0;
		cout << "No f contrasts" << endl;
	  }

	if(numFcontrasts > 0)
	  {
	    // Check contrasts match
	    if(tcontrasts.Nrows() != fcontrasts.Ncols())
	      { 
		cerr << "tcontrasts.Nrows()  = " << tcontrasts.Nrows() << endl;
		cerr << "fcontrasts.Ncols()  = " << fcontrasts.Ncols() << endl;
		throw Exception("size of tcontrasts does not match fcontrasts\n");
	      }
	    
if(numFcontrasts > 0)
	    setupfcontrasts();
	  }    	
      }    

    
  }

  void Design::setupfcontrasts()
  {
    Tracer_Plus trace("Design::setupfcontrasts");
    
    fc.resize(numFcontrasts);
    //reduceddms.resize(numFcontrasts);

    for(int f=0; f < numFcontrasts; f++)
    {
      fc[f].ReSize(tcontrasts.Nrows(),nevs);
      fc[f] = 0;

      int count = 0;
    
      for(int c = 1; c <= tcontrasts.Nrows(); c++)
	{	
	  if(fcontrasts(f+1,c) == 1)
	    {	    
	      count++;
	      fc[f].Row(count) = tcontrasts.Row(c);	

	    }
	}
    
      fc[f] = fc[f].Rows(1,count);

//       const Matrix& tfc = fc[f]; 
      
//       reduceddms[f] = getdm()*(Identity(nevs)-tfc*pinv(tfc));
      
//       Matrix tmp = zeros(2,2);
//       tmp(1,1) = 1;
//       reduceddms[f] = getdm()*(Identity(nevs)-tmp);

//        OUT(f);
//        OUT(fc[f]);
//       OUT(fc[f].t()*fc[f].i()*fc[f].t());	
//       OUT(dm);
//        OUT(reduceddms[f]);


    }
  }

  
}

