/*  gsoptions.h

    Mark Woolrich, Tim Behrens - FMRIB Image Analysis Group

    Copyright (C) 2002 University of Oxford  */

/*  CCOPYRIGHT  */

#if !defined(GsOptions_h)
#define GsOptions_h

#include <string>
#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <stdio.h>
#include "utils/options.h"
#include "utils/log.h"

using namespace Utilities;

namespace Gs {

class GsOptions {
 public:
  static GsOptions& getInstance();
  ~GsOptions() { delete gopt; }
  
  Option<bool> verbose;
  Option<int> debuglevel;
  Option<bool> timingon;
  Option<bool> help;
  Option<string> copefile;
  Option<string> varcopefile;
  Option<string> dofvarcopefile;
  Option<string> maskfile;
  Option<string> designfile;
  Option<string> covsplitfile;
  Option<string> tcontrastsfile;
  Option<string> fcontrastsfile;
  Option<string> logdir;
  Option<int> njumps;
  Option<int> burnin;
  Option<int> sampleevery;
  Option<bool> fixed;
  Option<float> zlowerthreshold;
  Option<float> zupperthreshold;
  Option<bool> fixmeanfortfit;
  Option<bool> justols;
  Option<bool> ols;
  Option<bool> em;

  void parse_command_line(int argc, char** argv, Log& logger);
  
 private:
  GsOptions();  
  const GsOptions& operator=(GsOptions&);
  GsOptions(GsOptions&);

  OptionParser options; 
      
  static GsOptions* gopt;
  
};

 inline GsOptions& GsOptions::getInstance(){
   if(gopt == NULL)
     gopt = new GsOptions();
   
   return *gopt;
 }

 inline GsOptions::GsOptions() :
   verbose(string("-v,-V,--verbose"), false, 
	   string("switch on diagnostic messages"), 
	   false, no_argument),
   debuglevel(string("--debug,--debuglevel"), 0,
		       string("set debug level"), 
		       false, requires_argument),
   timingon(string("--to,--timingon"), false, 
		       string("turn timing on"), 
		       false, no_argument),
   help(string("-h,--help"), false,
		    string("display this message"),
		    false, no_argument),
   copefile(string("--cope,--copefile"), string("cope"),
			  string("cope regressor data file"),
		     true, requires_argument),  
   varcopefile(string("--vc,--vcope,--varcope,--varcopefile"), string("varcope"),
			  string("varcope weightings data file"),
		     true, requires_argument),
   dofvarcopefile(string("--dvc,--dvcope,--dofvarcope,--dofvarcopefile"), string(""),
			  string("DOF data file for the varcope data"),
	    false, requires_argument),
   maskfile(string("--mask,--maskfile"), string(""),
			  string("mask file"),
		     true, requires_argument),
   designfile(string("--dm,--designfile"), string("design.mat"),
			  string("design matrix file"),
		     true, requires_argument), 
   covsplitfile(string("--cs,--csf,--covsplitfile"), string("design.gs"),
			  string("file containing an ASCII matrix specifying the groups the covariance is split into"),
		     true, requires_argument),
   tcontrastsfile(string("--tc,--tcf,--tcontrastsfile"), string("design.con"),
		  string("file containing an ASCII matrix specifying the t contrasts"),
		  true, requires_argument),  
   fcontrastsfile(string("--fc,--fcf,--fcontrastsfile"), string(""),
		  string("file containing an ASCII matrix specifying the f contrasts"),
		  false, requires_argument),  
   logdir(string("--ld,--logdir"), string("logdir"),
			  string("log directory"),
		     false, requires_argument),  
   njumps(string("--nj,--njumps"), 5000,
			  string("Num of jumps to be made by MCMC"),
		     false, requires_argument),
   burnin(string("--bi,--burnin"), 500,
			  string("Num of jumps at start of MCMC to be discarded"),
		     false, requires_argument),
   sampleevery(string("--se,--sampleevery"), 1,
			  string("Num of jumps for each sample"),
		     false, requires_argument),
   fixed(string("--fixed"), false, 
                      string("Fixed effects"), 
		      false, no_argument),
   zlowerthreshold(string("--zlt"), 0,
			  string("Absolute value of lower threshold on post-OLS, pre-MCMC z-stats"),
		     false, requires_argument),
   zupperthreshold(string("--zut"), 0,
			  string("Absolute value of upper threshold on post-OLS, pre-MCMC z-stats"),
		     false, requires_argument),
   fixmeanfortfit(string("--fm,--fixmean"), false, 
                      string("fix mean for tfit"), 
		      false, no_argument),
   justols(string("--jols"), false, 
                      string("do just OLS"), 
		      false, no_argument),
   ols(string("--ols"), false, 
                      string("do OLS"), 
		      false, no_argument),
   em(string("--em"), false, 
      string("do em (default is strapon)"), 
      false, no_argument),
   options("flame", "flame --verbose\n")
   {
     try {
       options.add(verbose);
       options.add(debuglevel);
       options.add(timingon);
       options.add(help);
       options.add(copefile);
       options.add(varcopefile);
       options.add(dofvarcopefile);
       options.add(maskfile);
       options.add(designfile);
       options.add(tcontrastsfile);
       options.add(fcontrastsfile);
       options.add(covsplitfile);
       options.add(logdir);
       options.add(njumps);
       options.add(burnin);
       options.add(sampleevery);
       options.add(fixed);
       options.add(zlowerthreshold);
       options.add(zupperthreshold);
       options.add(fixmeanfortfit);
       options.add(justols);
       options.add(ols);
       options.add(em);
     }
     catch(X_OptionError& e) {
       options.usage();
       cerr << endl << e.what() << endl;
     } 
     catch(std::exception &e) {
       cerr << e.what() << endl;
     }    
     
   }
}

#endif



