/*  applywarp.cc

    Mark Jenkinson, FMRIB Image Analysis Group

    Copyright (C) 2001 University of Oxford  */

/*  CCOPYRIGHT  */

#include "utils/options.h"
#include "miscmaths/miscmaths.h"
#include "warpfns.h"

#define _GNU_SOURCE 1
#define POSIX_SOURCE 1

using namespace Utilities;
using namespace NEWMAT;
using namespace MISCMATHS;

////////////////////////////////////////////////////////////////////////////

// COMMAND LINE OPTIONS

string title="applywarp (Version 1.2)\nCopyright(c) 2001, University of Oxford (Mark Jenkinson)";
string examples="applywarp -i invol -o outvol -r refvol -w warpvol";

Option<bool> verbose(string("-v,--verbose"), false, 
		     string("switch on diagnostic messages"), 
		     false, no_argument);
Option<bool> help(string("-h,--help"), false,
		  string("display this message"),
		  false, no_argument);
Option<string> interp(string("--interp"), string(""),
		   string("interpolation method {nn,trilinear,sinc}"),
		   false, requires_argument);
Option<string> inname(string("-i,--in"), string(""),
		       string("filename of input image (to be warped)"),
		       true, requires_argument);
Option<string> refname(string("-r,--ref"), string(""),
		       string("filename for reference image"),
		       false, requires_argument);
Option<string> outname(string("-o,--out"), string(""),
		       string("filename for output (warped) image"),
		       true, requires_argument);
Option<string> warpname(string("-w,--warp"), string(""),
			string("filename for warp transform (volume)"),
			true, requires_argument);
Option<string> prematname(string("--premat"), string(""),
		       string("filename for pre-transform (affine matrix)"),
		       false, requires_argument);
Option<string> postmatname(string("--postmat"), string(""),
		       string("filename for post-transform (affine matrix)"),
		       false, requires_argument);


////////////////////////////////////////////////////////////////////////////

int applywarp()
{

  // read in pre/post transforms
  Matrix premat, postmat;
  premat = Identity(4);
  postmat = Identity(4);
  if (prematname.set()) {
    read_ascii_matrix(premat,prematname.value());
  }
  if (postmatname.set()) {
    read_ascii_matrix(postmat,postmatname.value());
  }
  
  // read in images
  volume4D<float> invol, outvol;
  volume<float> refvol;
  volumeinfo vinfo;
  volume4D<float> warpvol;

  read_volume4D(invol,inname.value(),vinfo);
  read_volume4D(warpvol,warpname.value());
  if (warpvol.tsize()<3) {
    cerr << "Warp volume does not have the correct dimensions (min. 3 volumes)"
	 << endl;
    exit(1);
  }
  if (refname.set()) {
    read_volume(refvol,refname.value());
  } else {
    refvol = warpvol[0];
  }
  outvol = invol;

  volume<float> tmpvol;
  for (int t=0; t<invol.tsize(); t++) {
    invol[t].setpadvalue(invol[t].backgroundval());
    invol[t].setextrapolationmethod(extraslice);
    
    // set interpolation method
    if (interp.value() == "nn" ) {
      invol.setinterpolationmethod(nearestneighbour);
    } else if (interp.value() == "trilinear") {
      invol.setinterpolationmethod(trilinear);
    } else if (interp.value() == "sinc") {
      invol.setinterpolationmethod(sinc);
    }
    
    // do the deed
    tmpvol = refvol;
    apply_warp(invol[t],tmpvol,warpvol,premat,postmat);
    outvol[t] = tmpvol;
  }

  // save the results
  save_volume4D_dtype(outvol,outname.value(),dtype(inname.value()),vinfo);
  
  return 0;
}




int main(int argc, char *argv[])
{

  Tracer tr("main");

  OptionParser options(title, examples);

  try {
    options.add(inname);
    options.add(refname);
    options.add(warpname);
    options.add(outname);
    options.add(prematname);
    options.add(postmatname);
    options.add(interp);
    options.add(verbose);
    options.add(help);
    
    options.parse_command_line(argc, argv);

    if ( (help.value()) || (!options.check_compulsory_arguments(true)) )
      {
	options.usage();
	exit(EXIT_FAILURE);
      }
  }  catch(X_OptionError& e) {
    options.usage();
    cerr << endl << e.what() << endl;
    exit(EXIT_FAILURE);
  } catch(std::exception &e) {
    cerr << e.what() << endl;
  } 

  return applywarp();
}

