/*
  irvanc - fMRI IRVA analysis of non-cyclic data

  Stuart Clare, FMRIB Physics Group

  Copyright (c) 1998-2000 University of Oxford
   
  Clare, S. et.al. (1999) Magnetic Resonance in Medicine. 42,1117-1122.
*/

/* CCOPYRIGHT */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "avwio/avwio.h"
#include "ftoz.h"

/* usage: irvanc infile Foutfile Zoutfile %threshold events eventsize eventfile [-av filename]*/

#define MAXLEN 128

int main(int argc,char *argv[])
{
  char infile[MAXLEN],foutfile[MAXLEN],zoutfile[MAXLEN];
  char trigfile[MAXLEN],avfile[MAXLEN];
  short *image,thold;
  short max=0;
  double *F,*Z,thresh;
  int h,i,j,k,l,points,cycles,size,av=0;
  short ppl,lpi,ipv,vols,type,*trigs;
  long volsize,c;
  float *fdata,*average=NULL;
  double *data,T,C,SST,SSTr,Ti,SSE;
  FILE *fp;
  AVW *ip,*op;

  if(argc<8){
    printf("usage: irvanc infile Ffile Zfile %%thold events eventsize eventfile [-av file]\n");
    return(1);
  }

  sprintf(infile,"%s",argv[1]);
  sprintf(foutfile,"%s",argv[2]);
  sprintf(zoutfile,"%s",argv[3]);
  sscanf(argv[4],"%lf",&thresh);
  sscanf(argv[5],"%d",&cycles);
  sscanf(argv[6],"%d",&points);
  sscanf(argv[7],"%s",trigfile);
  if(argc>8){
    if(!strcmp(argv[8],"-av")){
      av=1;
      sprintf(avfile,"%s",argv[9]);
    }
  }

  /* Open input AVW file */
  if((ip=AvwOpen(infile,"r"))==NULL){
    printf("Failed to open AVW file: %s\n",infile);
    return(1);
  }
  AvwGetDim(ip,&ppl,&lpi,&ipv,&vols);
  AvwGetDataType(ip,&type);
  
  if(type!=DT_SIGNED_SHORT){
    printf("Data type must be signed short");
    return(1);
  }
  volsize=ppl*lpi*ipv;

  if((trigs=(short *)malloc(cycles*sizeof(short)))==NULL){
    printf("Malloc failed.");
    return(1);
  }
  size=points*cycles;
  if(size>vols){
    printf("Events times event size greater than number of volumes");
    return(1);
  }

  if((fp=fopen(trigfile,"r"))==NULL){
    printf("Can't open: %s",trigfile);
    return(1);
  }
  for(i=0;i<cycles;i++){
    fscanf(fp,"%hd\n",&trigs[i]);
    if((trigs[i]+points)>vols){
      printf("Not enough images for event file");
      return(1);
    }
  }
  /* Malloc */
  if(volsize>ppl*vols){
    if((image=(short *)malloc(volsize*sizeof(short)))==NULL){
      printf("Malloc failed.\n");
      return(1);
    }
  } else {
    if((image=(short *)malloc(ppl*vols*sizeof(short)))==NULL){
      printf("Malloc failed.\n");
      return(1);
    }
  }
  if((data=(double *)malloc(vols*sizeof(double)))==NULL){
    printf("Malloc failed.\n");
    return(1);
  }
  if((F=(double *)malloc(volsize*sizeof(double)))==NULL){
    printf("Malloc failed.\n");
    return(1);
  }
  if((Z=(double *)malloc(volsize*sizeof(double)))==NULL){
    printf("Malloc failed.\n");
    return(1);
  }
  if((fdata=(float *)malloc(volsize*sizeof(float)))==NULL){
    printf("Malloc failed.\n");
    return(1);
  }
  if(av){
    if((average=(float *)malloc(points*volsize*sizeof(float)))==NULL){
      printf("Malloc failed.\n");
      return(1);
    }
  }

  /* Read in first image and find max */
  AvwSeekVolume(ip,0);
  if(AvwReadVolumes(ip,image,1)!=1){
    printf("Read error: %s\n",infile);
    return(1);
  }
  for(i=0;i<volsize;i++)if(image[i]>max)max=image[i];
  thold=(short)(max*thresh/100);

  /* For each line in image */
  for(h=0;h<ipv;h++){
    for(i=0;i<lpi;i++){
      /* Read in data */	
      for(k=0;k<vols;k++){
	AvwSeekVolume(ip,k);
	if(AvwReadRowSeries(ip,&image[k*ppl],i,h,1)!=1){
	  printf("Read error: %s\n",infile);
	  return(1);
	}
      }

      /* For each point in line */
      for(j=0;j<ppl;j++){
	if(image[j]>thold){
	  for(k=0;k<cycles;k++){
	    for(l=0;l<points;l++){
	      data[(k*points)+l]=(double)image[((trigs[k]+l)*ppl)+j];
	    }
	  }

	  T=SST=0.0;
	  for(k=0;k<size;k++){
	    T+=data[k];
	    SST+=(data[k]*data[k]);
	  }
	  C=(T*T)/size;
	  SST-=C;
	  SSTr=0.0;
	  for(k=0;k<points;k++){
	    for(l=0,Ti=0.0;l<cycles;l++){
	      Ti+=data[(l*points)+k];
	    }
	    SSTr+=(Ti*Ti);
	  }
	  SSTr/=cycles;
	  SSTr-=C;
	  SSE=SST-SSTr;
	  F[(h*lpi*ppl)+(i*ppl)+j]=(SSTr/(points-1))/(SSE/(points*(cycles-1)));
	}
	else F[(h*lpi*ppl)+(i*ppl)+j]=0;
      }
    }
  }

  /* Create average image */
  if(av){
    if((op=AvwOpen(avfile,"w"))==NULL){
      printf("Unable to create AVW file: %s\n",avfile);
    }
    AvwCloneHeader(op,ip);
    AvwSetDim(op,ppl,lpi,ipv,points);
    AvwSetDataType(op,DT_FLOAT);
    AvwSetMinMax(op,0,max);

    for(c=0;c<volsize*points;c++)average[c]=0.0;
    for(j=0;j<cycles;j++){
      AvwSeekVolume(ip,trigs[j]);
      for(k=0;k<points;k++){
	if(AvwReadVolumes(ip,image,1)!=1){
	  printf("Read error: %s\n",infile);
	  return(1);
	}
	for(c=0;c<volsize;c++)
	  average[k*volsize+c]+=((float)image[c]/(float)cycles);
      }
    }
    AvwSeekVolume(op,0);
    if((short)AvwWriteVolumes(op,average,points)!=points){
      printf("Write error: %s\n",avfile);
      return(1);
    }
    AvwClose(op);
  }

  /* Transform F to Z */
  ftoz(F,Z,volsize,points-1,points*(cycles-1));
  
  /* Write output */
  if((op=AvwOpen(foutfile,"w"))==NULL){
    printf("Unable to create AVW file: %s\n",foutfile);
    return(1);
  }
  AvwCloneHeader(op,ip);
  AvwSetDim(op,ppl,lpi,ipv,1);
  AvwSetDataType(op,DT_FLOAT);
  AvwSetMinMax(op,0,10);
  for(c=0;c<volsize;c++)fdata[c]=(float)F[c];
  AvwSeekVolume(op,0);
  if(AvwWriteVolumes(op,fdata,1)!=1){
    printf("Write error: %s\n",foutfile);
    return(1);
  }
  AvwClose(op);
  
  if((op=AvwOpen(zoutfile,"w"))==NULL){
    printf("Unable to create AVW file: %s\n",zoutfile);
    return(1);
  }
  AvwCloneHeader(op,ip);
  AvwSetDim(op,ppl,lpi,ipv,1);
  AvwSetDataType(op,DT_FLOAT);
  AvwSetMinMax(op,0,10);
  for(c=0;c<volsize;c++)fdata[c]=(float)Z[c];
  AvwSeekVolume(op,0);
  if(AvwWriteVolumes(op,fdata,1)!=1){
    printf("Write error: %s\n",foutfile);
    return(1);
  }
  AvwClose(op);
  AvwClose(ip);

  return(0);
}
